#include "XOPStandardHeaders.h"			// Include ANSI headers, Mac headers, IgorXOP.h, XOP.h and XOPSupport.h
#include "igorinterface.h"

namespace  {

  #pragma pack(2)	// All structures passed to Igor are two-byte aligned.
  struct runningInMainThreadParams{
	  UserFunctionThreadInfoPtr tp; // needed for thread safe functions
	  double  result;
  };
  typedef struct runningInMainThreadParams runningInMainThreadParams;
  #pragma pack()

  extern "C" int runningInMainThread(runningInMainThreadParams *p);

  #pragma pack(2)	// All structures passed to Igor are two-byte aligned.
  struct threadSafeWrapperParams{
	  Handle  function;
	  UserFunctionThreadInfoPtr tp; // needed for thread safe functions
	  Handle  result;
  };
  typedef struct threadSafeWrapperParams threadSafeWrapperParams;
  #pragma pack()

  extern "C" int threadSafeWrapper(threadSafeWrapperParams *p);

  XOPIORecResult RegisterFunction()
  {
    /*	NOTE:
        Some XOPs should return a result of NIL in response to the FUNCADDRS message.
        See XOP manual "Restrictions on Direct XFUNCs" section.
    */
    const int funcIndex = GetXOPItem(0);		/* which function invoked ? */

    switch (funcIndex)
    {
    case 0:						
      return (XOPIORecResult) runningInMainThread;
      break;
    case 1:						
      return (XOPIORecResult) threadSafeWrapper;
      break;
    }
    return NIL;
  }

  const int FORCE_GLOBAL = 1;

} // anonymous namespace 

extern "C" int runningInMainThread(runningInMainThreadParams *p)
{
	p->result = RunningInMainThread();
	return 0;
}

extern "C" int threadSafeWrapper(threadSafeWrapperParams *p)
{
  int ret;
  p->result = NewHandle(0LL);

  if( !RunningInMainThread() )
    return 0;

  char function[MAXCMDLEN + 1];
  if( ret = GetCStringFromHandle(p->function,function,MAXCMDLEN) )
    return ret;

  FunctionInfo fip;
  if( ret = GetFunctionInfo(function, &fip) )
    return ret;

  if ( fip.numRequiredParameters != 0 || fip.returnType != HSTRING_TYPE ) 
    return WRONG_FUNC_PARAMS;

  const char baseName[] = "threadSafeWrapperVar";
  char resultStorageName[MAX_OBJ_NAME + 1];
  int suffix = 0;

  if( ret = UniqueName2(MAIN_NAME_SPACE, baseName, resultStorageName, &suffix) )
    return ret;

  if( ret = SetIgorStringVar(resultStorageName, "", FORCE_GLOBAL) )
  {
    KillDataFolderObject(NULL, STR_OBJECT, resultStorageName);
    return ret;
  }

  char cmdToExecute[MAXCMDLEN + 1];
  const int len = _snprintf(cmdToExecute,MAXCMDLEN,"%s = %s()",resultStorageName, function);
  cmdToExecute[MAXCMDLEN] = '\0';
  if( len < 0 || len > MAXCMDLEN )
  {
    KillDataFolderObject(NULL, STR_OBJECT, resultStorageName);
    return LINE_TOO_LONG;
  }

  if( ret = XOPSilentCommand(cmdToExecute) )
  {
    KillDataFolderObject(NULL, STR_OBJECT, resultStorageName);
    return ret;
  }

  const Handle resultHandle = FetchStrHandle(resultStorageName);

  if( resultHandle == NULL
      || ( ret = HandAndHand(resultHandle, p->result) )
    )
  {
    KillDataFolderObject(NULL, STR_OBJECT, resultStorageName);
    return ret;
  }

  if( ret = KillDataFolderObject(NULL, STR_OBJECT, resultStorageName) )
    return ret;

  return 0;
}

/*	XOPEntry()

	This is the entry point from the host application to the XOP for all messages after the
	INIT message.
*/

extern "C" void XOPEntry(void)
{	
	XOPIORecResult result = 0;

	switch (GetXOPMessage())
  {
  	case FUNCADDRS:
			result = RegisterFunction();
			break;
	}
	SetXOPResult(result);
}

/*	main(ioRecHandle)

	This is the initial entry point at which the host application calls XOP.
	The message sent by the host must be INIT.
	main() does any necessary initialization and then sets the XOPEntry field of the
	ioRecHandle to the address to be called for future messages.
*/

HOST_IMPORT int main(IORecHandle ioRecHandle)
{	
	XOPInit(ioRecHandle);					// Do standard XOP initialization
	SetXOPEntry(XOPEntry);					// Set entry point for future calls
	
	if (igorVersion < 620)
  {
		SetXOPResult(OLD_IGOR);
		return EXIT_FAILURE;
	}
	
	SetXOPResult(0);
	return EXIT_SUCCESS;
}
